@@ -37,7 +37,7 @@
 char *
 strstr(const char *s, const char *t)
 {
-    char *p1, *p2;
+    const char *p1, *p2;
 
     for (; *s; s++) {
         for (p1 = s, p2 = t; *p2; p1++, p2++)
@@ -187,40 +187,45 @@ zpathmax(char *dir)
 #endif /* 0 */
 
 #ifdef HAVE_SYSCONF
-/* This is replaced by a macro from system.h if not HAVE_SYSCONF.    *
- * 0 is returned by sysconf if _SC_OPEN_MAX is unavailable;          *
- * -1 is returned on error                                           *
- *                                                                   *
- * Neither of these should happen, but resort to OPEN_MAX rather     *
- * than return 0 or -1 just in case.                                 */
+/*
+ * This is replaced by a macro from system.h if not HAVE_SYSCONF.
+ * 0 is returned by sysconf if _SC_OPEN_MAX is unavailable;
+ * -1 is returned on error
+ *
+ * Neither of these should happen, but resort to OPEN_MAX rather
+ * than return 0 or -1 just in case.
+ *
+ * We'll limit the open maximum to ZSH_INITIAL_OPEN_MAX to
+ * avoid probing ridiculous numbers of file descriptors.
+ */
 
 /**/
 mod_export long
 zopenmax(void)
 {
-    static long openmax = 0;
+    long openmax;
 
-    if (openmax < 1) {
-	if ((openmax = sysconf(_SC_OPEN_MAX)) < 1) {
-	    openmax = OPEN_MAX;
-	} else if (openmax > OPEN_MAX) {
-	    /* On some systems, "limit descriptors unlimited" or the  *
-	     * equivalent will set openmax to a huge number.  Unless  *
-	     * there actually is a file descriptor > OPEN_MAX already *
-	     * open, nothing in zsh requires the true maximum, and in *
-	     * fact it causes inefficiency elsewhere if we report it. *
-	     * So, report the maximum of OPEN_MAX or the largest open *
-	     * descriptor (is there a better way to find that?).      */
-	    long i, j = OPEN_MAX;
-	    for (i = j; i < openmax; i += (errno != EINTR)) {
-		errno = 0;
-		if (fcntl(i, F_GETFL, 0) < 0 &&
-		    (errno == EBADF || errno == EINTR))
-		    continue;
-		j = i;
-	    }
-	    openmax = j;
+    if ((openmax = sysconf(_SC_OPEN_MAX)) < 1) {
+	openmax = OPEN_MAX;
+    } else if (openmax > OPEN_MAX) {
+	/* On some systems, "limit descriptors unlimited" or the  *
+	 * equivalent will set openmax to a huge number.  Unless  *
+	 * there actually is a file descriptor > OPEN_MAX already *
+	 * open, nothing in zsh requires the true maximum, and in *
+	 * fact it causes inefficiency elsewhere if we report it. *
+	 * So, report the maximum of OPEN_MAX or the largest open *
+	 * descriptor (is there a better way to find that?).      */
+	long i, j = OPEN_MAX;
+	if (openmax > ZSH_INITIAL_OPEN_MAX)
+	    openmax = ZSH_INITIAL_OPEN_MAX;
+	for (i = j; i < openmax; i += (errno != EINTR)) {
+	    errno = 0;
+	    if (fcntl(i, F_GETFL, 0) < 0 &&
+		(errno == EBADF || errno == EINTR))
+		continue;
+	    j = i;
 	}
+	openmax = j;
     }
 
     return (max_zsh_fd > openmax) ? max_zsh_fd : openmax;
@@ -270,11 +275,7 @@ zgetdir(struct dirsav *d)
     buf[pos] = '\0';
     strcpy(nbuf, "../");
     if (stat(".", &sbuf) < 0) {
-	if (d)
-	    return NULL;
-	buf[0] = '.';
-	buf[1] = '\0';
-	return buf;
+	return NULL;
     }
 
     /* Record the initial inode and device */
@@ -349,6 +350,11 @@ zgetdir(struct dirsav *d)
 	closedir(dir);
 	if (!de)
 	    break;		/* Not found */
+	/*
+	 * We get the "/" free just by copying from nbuf+2 instead
+	 * of nbuf+3, which is where we copied the path component.
+	 * This means buf[pos] is always a "/".
+	 */
 	len = strlen(nbuf + 2);
 	pos -= len;
 	while (pos <= 1) {
@@ -366,15 +372,22 @@ zgetdir(struct dirsav *d)
 	if (chdir(".."))
 	    break;
     }
+
+    /*
+     * Fix up the directory, if necessary.
+     * We're changing back down the hierarchy, ignore the
+     * "/" at buf[pos].
+     */
     if (d) {
 #ifndef HAVE_FCHDIR
-	if (*buf)
+	if (buf[pos])
 	    zchdir(buf + pos + 1);
 	noholdintr();
 #endif
 	return NULL;
     }
-    if (*buf)
+
+    if (buf[pos])
 	zchdir(buf + pos + 1);
     noholdintr();
 
@@ -392,20 +405,62 @@ zgetdir(struct dirsav *d)
     }
 #endif
 
-    buf[0] = '.';
-    buf[1] = '\0';
-    return buf;
+    /*
+     * Something bad happened.
+     * This has been seen when inside a special directory,
+     * such as the Netapp .snapshot directory, that doesn't
+     * appear as a directory entry in the parent directory.
+     * We'll just need our best guess.
+     *
+     * We only get here from zgetcwd(); let that fall back to pwd.
+     */
+
+    return NULL;
 }
 
+/*
+ * Try to find the current directory.
+ * If we couldn't work it out internally, fall back to getcwd().
+ * If it fails, fall back to pwd; if zgetcwd() is being used
+ * to set pwd, pwd should be NULL and we just return ".".
+ */
+
 /**/
 char *
 zgetcwd(void)
 {
-    return zgetdir(NULL);
+    char *ret = zgetdir(NULL);
+#ifdef HAVE_GETCWD
+    if (!ret) {
+#ifdef GETCWD_CALLS_MALLOC
+	char *cwd = getcwd(NULL, 0);
+	if (cwd) {
+	    ret = dupstring(cwd);
+	    free(cwd);
+	}
+#else
+	char *cwdbuf = zalloc(PATH_MAX);
+	ret = getcwd(cwdbuf, PATH_MAX);
+	if (ret)
+	    ret = dupstring(ret);
+	zfree(cwdbuf, PATH_MAX);
+#endif /* GETCWD_CALLS_MALLOC */
+    }
+#endif /* HAVE_GETCWD */
+    if (!ret)
+	ret = pwd;
+    if (!ret)
+	ret = dupstring(".");
+    return ret;
 }
 
-/* chdir with arbitrary long pathname.  Returns 0 on success, -1 on normal *
- * failure and -2 when chdir failed and the current directory is lost.  */
+/*
+ * chdir with arbitrary long pathname.  Returns 0 on success, -1 on normal *
+ * failure and -2 when chdir failed and the current directory is lost.
+ *
+ * This is to be treated as if at system level, so dir is unmetafied but
+ * terminated by a NULL.
+ */
 
 /**/
 mod_export int
@@ -580,7 +635,7 @@ strtoul(nptr, endptr, base)
 #endif /* HAVE_STRTOUL */
 
 /**/
-#ifdef BROKEN_WCWIDTH
+#if defined(BROKEN_WCWIDTH) && (defined(__STDC_ISO_10646__) || defined(__APPLE__))
 
 /*
  * This is an implementation of wcwidth() and wcswidth() (defined in
@@ -771,7 +826,7 @@ mk_wcwidth(wchar_t ucs)
 
   /* if we arrive here, ucs is not a combining or C0/C1 control character */
 
-  return 1 + 
+  return 1 +
     (ucs >= 0x1100 &&
      (ucs <= 0x115f ||                    /* Hangul Jamo init. consonants */
       ucs == 0x2329 || ucs == 0x232a ||
@@ -899,5 +954,20 @@ int mk_wcswidth_cjk(const wchar_t *pwcs, size_t n)
 #endif /* 0 */
 
 /**/
-#endif /* BROKEN_WCWIDTH */
+#endif /* BROKEN_WCWIDTH && (__STDC_ISO_10646__ || __APPLE__) */
 
+/**/
+#if defined(__APPLE__) && defined(BROKEN_ISPRINT)
+
+/**/
+int
+isprint_ascii(int c)
+{
+    if (!strcmp(nl_langinfo(CODESET), "UTF-8"))
+	return (c >= 0x20 && c <= 0x7e);
+    else
+	return isprint(c);
+}
+
+/**/
+#endif /* __APPLE__ && BROKEN_ISPRINT */
